from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING, Any, List, Optional, Tuple

from great_expectations.core.data_context_key import DataContextVariableKey
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.data_context_variables import (
    DataContextVariableSchema,
)
from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.data_context.types.base import DataContextConfig
from great_expectations.exceptions.exceptions import StoreBackendError
from great_expectations.util import filter_properties_dict

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.file_data_context import (
        FileDataContext,
    )


logger = logging.getLogger(__name__)

yaml = YAMLHandler()


class InlineStoreBackend(StoreBackend):
    """
    The InlineStoreBackend enables CRUD behavior with the fields noted in a user's project config (`great_expectations.yml`).

    The primary value of the InlineStoreBackend is the ability to modify either the entire config or very granular parts of it through the
    same interface. Whether it be replacing the entire config with a new one or tweaking an individual datasource nested within the config,
    a user of the backend is able to do so through the same key structure.

    For example:
        ("data_context", "")             -> Key used to get/set an entire config
        ("datasources", "my_datasource") -> Key used to get/set a specific datasource named "my_datasource"

    It performs these actions through a reference to a DataContext instance.
    Please note that is it only to be used with file-backed DataContexts (DataContext and FileDataContext).
    """

    def __init__(
        self,
        data_context: FileDataContext,
        resource_type: DataContextVariableSchema,
        runtime_environment: Optional[dict] = None,
        fixed_length_key: bool = False,
        suppress_store_backend_id: bool = False,
        manually_initialize_store_backend_id: str = "",
        store_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            store_name=store_name,
        )

        self._data_context = data_context
        self._resource_type = resource_type

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.
        self._config = {
            "runtime_environment": runtime_environment,
            "fixed_length_key": fixed_length_key,
            "suppress_store_backend_id": suppress_store_backend_id,
            "manually_initialize_store_backend_id": manually_initialize_store_backend_id,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    @property
    def config(self) -> dict:
        return self._config

    def _get(self, key: Tuple[str, ...]) -> Any:
        resource_name = InlineStoreBackend._determine_resource_name(key)
        project_config: DataContextConfig = self._data_context.config
        resource_type = self._resource_type

        if resource_type is DataContextVariableSchema.ALL_VARIABLES:
            return project_config

        variable_config: Any = project_config[resource_type]

        if resource_name is not None:
            return variable_config[resource_name]

        return variable_config

    def _set(self, key: Tuple[str, ...], value: Any, **kwargs: dict) -> None:
        resource_name = InlineStoreBackend._determine_resource_name(key)
        project_config: DataContextConfig = self._data_context.config
        resource_type = self._resource_type

        if resource_type is DataContextVariableSchema.ALL_VARIABLES:
            config_commented_map_from_yaml = yaml.load(value)
            value = DataContextConfig.from_commented_map(
                commented_map=config_commented_map_from_yaml
            )
            self._data_context.set_config(value)
        elif resource_name is not None:
            project_config[resource_type][resource_name] = value
        else:
            project_config[resource_type] = value

        self._save_changes()

    def _move(
        self, source_key: Tuple[str, ...], dest_key: Tuple[str, ...], **kwargs: dict
    ) -> None:
        raise StoreBackendError(
            "InlineStoreBackend does not support moving of keys; the DataContext's config variables schema is immutable"
        )

    def list_keys(self, prefix: Tuple[str, ...] = ()) -> List[str]:
        """
        See `StoreBackend.list_keys` for more information.

        Args:
            prefix: If supplied, allows for a more granular listing of nested values within the config.
                    Example: prefix=(datasources,) will list all datasource configs instead of top level keys.

        Returns:
            A list of string keys from the user's project config.
        """
        config_section: Optional[str] = None
        if prefix:
            config_section = prefix[0]

        keys: List[str]
        config_dict: dict = self._data_context.config.to_dict()
        if config_section is None:
            keys = list(key for key in config_dict.keys())
        else:
            config_values: dict = config_dict[config_section]
            if not isinstance(config_values, dict):
                raise StoreBackendError(
                    "Cannot list keys in a non-iterable section of a project config"
                )
            keys = list(key for key in config_values.keys())

        return keys

    def remove_key(self, key: Tuple[str, ...]) -> None:
        """
        See `StoreBackend.remove_key` for more information.
        """
        resource_name = InlineStoreBackend._determine_resource_name(key)
        resource_type = self._resource_type

        if resource_type is DataContextVariableSchema.ALL_VARIABLES:
            raise StoreBackendError(
                "InlineStoreBackend does not support the deletion of the overall DataContext project config"
            )
        if resource_name is None:
            raise StoreBackendError(
                "InlineStoreBackend does not support the deletion of top level keys; the DataContext's config variables schema is immutable"
            )
        elif not self._has_key(key):
            raise StoreBackendError(
                f"Could not find a value associated with key `{key}`"
            )

        del self._data_context.config[resource_type][resource_name]

        self._save_changes()

    def build_key(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> DataContextVariableKey:
        """Get the store backend specific implementation of the key. id included for super class compatibility."""
        return DataContextVariableKey(
            resource_name=name,
        )

    def _has_key(self, key: Tuple[str, ...]) -> bool:
        resource_name = InlineStoreBackend._determine_resource_name(key)
        resource_type = self._resource_type

        if resource_name is not None:
            res: dict = self._data_context.config.get(resource_type) or {}
            return resource_name in res

        return resource_type in self._data_context.config

    def _save_changes(self) -> None:
        context = self._data_context
        config_filepath = pathlib.Path(context.root_directory) / context.GX_YML  # type: ignore[arg-type]

        try:
            with open(config_filepath, "w") as outfile:
                context.config.to_yaml(outfile)
        # In environments where wrting to disk is not allowed, it is impossible to
        # save changes. As such, we log a warning but do not raise.
        except PermissionError as e:
            logger.warning(f"Could not save project config to disk: {e}")

    @staticmethod
    def _determine_resource_name(key: Tuple[str, ...]) -> Optional[str]:
        resource_name: Optional[str] = key[0] or None
        return resource_name



import jsonschema
import pytest

from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.profile.json_schema_profiler import JsonSchemaProfiler
from great_expectations.render.renderer_configuration import MetaNotesFormat


@pytest.fixture
def simple_schema():
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "first_name": {"type": "string"},
            "age": {"type": "integer"},
        },
    }


@pytest.fixture
def complex_flat_schema():
    """This includes some descriptions."""
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": "An address",
        "type": "object",
        "properties": {
            "post-office-box": {"type": "string"},
            "street-name": {"type": "string"},
            "street-number": {
                "type": "integer",
                "description": "Only the address number.",
            },
            "locality": {"type": "string"},
            "region": {"type": "string"},
            "postal-code": {"type": "string"},
            "country-name": {"type": "string"},
        },
        "required": ["locality", "region", "country-name"],
    }


@pytest.fixture
def boolean_types_schema():
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "active": {"type": "boolean"},
            "optional": {
                "anyOf": [
                    {
                        "type": "boolean",
                    },
                    {
                        "type": "null",
                    },
                ]
            },
        },
    }


@pytest.fixture
def enum_types_schema():
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "shirt-size": {"enum": ["XS", "S", "M", "XL", "XXL"]},
            "optional-color": {
                "anyOf": [
                    {
                        "enum": ["red", "green", "blue"],
                    },
                    {"type": "null"},
                ],
            },
            "optional-hat": {
                "type": ["string", "null"],
                "enum": ["red", "green", "blue"],
            },
            "optional-answer": {
                "enum": ["yes", "no", None],
            },
        },
    }


@pytest.fixture
def string_lengths_schema():
    """
    This fixture has various combinations string lengths.
    https://json-schema.org/understanding-json-schema/reference/string.html#length
    """
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "comments-no-constraints": {"type": "string"},
            "state-abbreviation-equal-min-max": {
                "type": "string",
                "minLength": 2,
                "maxLength": 2,
            },
            "ICD10-code-3-7": {"type": "string", "minLength": 3, "maxLength": 7},
            "name-no-max": {"type": "string", "minLength": 1},
            "password-max-33": {"type": "string", "maxLength": 33},
            "optional-min-1": {
                "anyOf": [
                    {
                        "type": "string",
                        "minLength": 1,
                    },
                    {
                        "type": "null",
                    },
                ]
            },
        },
    }


@pytest.fixture
def integer_ranges_schema():
    """
    This fixture has various combinations of integer ranges.
    https://json-schema.org/understanding-json-schema/reference/numeric.html#range
    """
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": "An address similar to http://microformats.org/wiki/h-card",
        "type": "object",
        "properties": {
            "favorite-number": {"type": "integer"},
            "age-0-130": {"type": "integer", "minimum": 0, "maximum": 130},
            "wheel-count-0-plus": {"type": "integer", "minimum": 0},
            "rpm-max-7000": {"type": "integer", "maximum": 7000},
            "lake-depth-max-minus-100": {"type": "integer", "maximum": -100},
            "floor-exclusive-min-0": {"type": "integer", "exclusiveMinimum": 0},
            "floor-exclusive-max-100": {"type": "integer", "exclusiveMaximum": 100},
            "gear-exclusive-0-6": {
                "type": "integer",
                "exclusiveMinimum": 0,
                "exclusiveMaximum": 6,
            },
            "optional-min-1": {
                "anyOf": [
                    {
                        "type": "integer",
                        "minimum": 1,
                    },
                    {
                        "type": "null",
                    },
                ]
            },
        },
    }


@pytest.fixture
def number_ranges_schema():
    """
    This fixture has various combinations of number ranges.
    https://json-schema.org/understanding-json-schema/reference/numeric.html#range
    """
    return {
        "$id": "https://example.com/address.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": "An address similar to http://microformats.org/wiki/h-card",
        "type": "object",
        "properties": {
            "favorite-number": {"type": "number"},
            "age-0-130": {"type": "number", "minimum": 0.5, "maximum": 130.5},
            "wheel-count-0-plus": {"type": "number", "minimum": 0.5},
            "rpm-max-7000": {"type": "number", "maximum": 7000.5},
            "lake-depth-max-minus-100": {"type": "number", "maximum": -100.5},
            "floor-exclusive-min-0": {"type": "number", "exclusiveMinimum": 0.5},
            "floor-exclusive-max-100": {"type": "number", "exclusiveMaximum": 100.5},
            "gear-exclusive-0-6": {
                "type": "number",
                "exclusiveMinimum": 0.5,
                "exclusiveMaximum": 6.5,
            },
            "optional-min-half": {
                "anyOf": [
                    {
                        "type": "number",
                        "minimum": 0.5,
                    },
                    {
                        "type": "null",
                    },
                ]
            },
        },
    }


@pytest.fixture
def null_fields_schema():
    """
    This fixture has null fields.
    https://json-schema.org/understanding-json-schema/reference/null.html
    """
    return {
        "$id": "https://example.com/null.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "null": {"type": "null"},
            "string-or-null": {"type": ["string", "null"]},
            "int-or-null": {"type": ["integer", "null"]},
            "number-or-null": {"type": ["number", "null"]},
            "enum-or-null": {"anyOf": [{"enum": ["a", "b", "c"]}, {"type": "null"}]},
        },
    }


def test_instantiable():
    profiler = JsonSchemaProfiler()
    assert isinstance(profiler, JsonSchemaProfiler)


def test_validate_returns_true_on_valid_schema(simple_schema):
    profiler = JsonSchemaProfiler()
    assert profiler.validate(simple_schema) is True


def test_profile_raises_errors_on_bad_inputs():
    profiler = JsonSchemaProfiler()
    for bad in [1, 1.1, None, "junk"]:
        with pytest.raises(TypeError):
            profiler.profile(bad, "foo")


def test_profile_raises_error_on_missing_suite_name(simple_schema):
    profiler = JsonSchemaProfiler()
    with pytest.raises(ValueError) as e:
        profiler.profile(simple_schema, suite_name=None)
    message = str(e.value)
    assert "provide a suite name" in message


def test_profile_raises_error_on_schema_missing_top_level_type_key():
    profiler = JsonSchemaProfiler()
    schema = {"a_schema": "missing_type"}
    with pytest.raises(KeyError) as e:
        profiler.profile(schema, "suite")
    message = str(e.value)
    assert "This profiler requires a json schema with a top level `type` key" in message


def test_profile_raises_error_on_schema_with_top_level_type_other_than_object():
    profiler = JsonSchemaProfiler()
    schema = {"type": "array"}
    with pytest.raises(TypeError) as e:
        profiler.profile(schema, "suite")
    message = str(e.value)
    assert (
        "This profiler requires a json schema with a top level `type` of `object`"
        in message
    )


def test_profile_enum_with_bad_input_raises_schema_error(enum_types_schema):
    profiler = JsonSchemaProfiler()
    # mangle the enum list
    enum_types_schema["properties"]["shirt-size"]["enum"] = "foo"
    with pytest.raises(jsonschema.SchemaError):
        profiler.profile(enum_types_schema, "enums")


def test_profile_simple_schema(empty_data_context, simple_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(simple_schema, "simple_suite")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "simple_suite"
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "first_name"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "first_name",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "first_name"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "age"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "age",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "age"},
            "meta": {},
        },
    ]
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_profile_boolean_schema(empty_data_context, boolean_types_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(boolean_types_schema, "bools")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "bools"
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "active"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "active",
                "type_list": [
                    "BIT",
                    "BOOL",
                    "BOOLEAN",
                    "BooleanType",
                    "BooleanType()",
                    "TINYINT",
                    "bool",
                    "boolean",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {"column": "active", "value_set": [True, False]},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "active"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "optional",
                "type_list": [
                    "BIT",
                    "BOOL",
                    "BOOLEAN",
                    "BooleanType",
                    "BooleanType()",
                    "TINYINT",
                    "bool",
                    "boolean",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {"column": "optional", "value_set": [True, False]},
            "meta": {},
        },
    ]
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_profile_enum_schema(empty_data_context, enum_types_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(enum_types_schema, "enums")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "enums"
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "shirt-size"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {
                "column": "shirt-size",
                "value_set": ["XS", "S", "M", "XL", "XXL"],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "shirt-size"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional-color"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {
                "column": "optional-color",
                "value_set": ["red", "green", "blue"],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional-hat"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "optional-hat",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {"column": "optional-hat", "value_set": ["red", "green", "blue"]},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional-answer"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "kwargs": {"column": "optional-answer", "value_set": ["yes", "no"]},
            "meta": {},
        },
    ]
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_profile_string_lengths_schema(empty_data_context, string_lengths_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(string_lengths_schema, "lengths")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "lengths"
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "comments-no-constraints"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "comments-no-constraints",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "comments-no-constraints"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "state-abbreviation-equal-min-max"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "state-abbreviation-equal-min-max",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_value_lengths_to_equal",
            "kwargs": {"column": "state-abbreviation-equal-min-max", "value": 2},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "state-abbreviation-equal-min-max"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "ICD10-code-3-7"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "ICD10-code-3-7",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_value_lengths_to_be_between",
            "kwargs": {"column": "ICD10-code-3-7", "min_value": 3, "max_value": 7},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "ICD10-code-3-7"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "name-no-max"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "name-no-max",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_value_lengths_to_be_between",
            "kwargs": {"column": "name-no-max", "min_value": 1},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "name-no-max"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "password-max-33"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "password-max-33",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_value_lengths_to_be_between",
            "kwargs": {"column": "password-max-33", "max_value": 33},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "password-max-33"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional-min-1"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "optional-min-1",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_value_lengths_to_be_between",
            "kwargs": {"column": "optional-min-1", "min_value": 1},
            "meta": {},
        },
    ]
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_profile_integer_ranges_schema(empty_data_context, integer_ranges_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(integer_ranges_schema, "integer_ranges")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "integer_ranges"
    # fmt: off
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "kwargs": {"column": "favorite-number"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "favorite-number",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "favorite-number"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "age-0-130"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "age-0-130",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "age-0-130", "min_value": 0, "max_value": 130},
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "age-0-130"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "wheel-count-0-plus"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "wheel-count-0-plus",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "wheel-count-0-plus", "min_value": 0},
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "wheel-count-0-plus"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "rpm-max-7000"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "rpm-max-7000",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "rpm-max-7000", "max_value": 7000},
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "rpm-max-7000"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "lake-depth-max-minus-100"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "lake-depth-max-minus-100",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "lake-depth-max-minus-100", "max_value": -100},
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "lake-depth-max-minus-100"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "floor-exclusive-min-0"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "floor-exclusive-min-0",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "floor-exclusive-min-0",
                "min_value": 0,
                "strict_min": True,
            },
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "floor-exclusive-min-0"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "floor-exclusive-max-100"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "floor-exclusive-max-100",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "floor-exclusive-max-100",
                "max_value": 100,
                "strict_max": True,
            },
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "floor-exclusive-max-100"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "gear-exclusive-0-6"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "gear-exclusive-0-6",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "gear-exclusive-0-6",
                "min_value": 0,
                "strict_min": True,
                "max_value": 6,
                "strict_max": True,
            },
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
        {
            "kwargs": {"column": "gear-exclusive-0-6"},
            "expectation_type": "expect_column_values_to_not_be_null",
            "meta": {},
        },
        {
            "kwargs": {"column": "optional-min-1"},
            "expectation_type": "expect_column_to_exist",
            "meta": {},
        },
        {
            "kwargs": {
                "column": "optional-min-1",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
        },
        {
            "kwargs": {"column": "optional-min-1", "min_value": 1},
            "expectation_type": "expect_column_values_to_be_between",
            "meta": {},
        },
    ]
    # fmt: on
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_profile_number_ranges_schema(empty_data_context, number_ranges_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(number_ranges_schema, "number_ranges")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "number_ranges"
    # fmt: off
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "favorite-number"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "favorite-number",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "favorite-number"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "age-0-130"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "age-0-130",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {"column": "age-0-130", "min_value": 0.5, "max_value": 130.5},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "age-0-130"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "wheel-count-0-plus"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "wheel-count-0-plus",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {"column": "wheel-count-0-plus", "min_value": 0.5},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "wheel-count-0-plus"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "rpm-max-7000"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "rpm-max-7000",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {"column": "rpm-max-7000", "max_value": 7000.5},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "rpm-max-7000"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "lake-depth-max-minus-100"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "lake-depth-max-minus-100",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {"column": "lake-depth-max-minus-100", "max_value": -100.5},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "lake-depth-max-minus-100"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "floor-exclusive-min-0"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "floor-exclusive-min-0",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {
                "column": "floor-exclusive-min-0",
                "min_value": 0.5,
                "strict_min": True,
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "floor-exclusive-min-0"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "floor-exclusive-max-100"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "floor-exclusive-max-100",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {
                "column": "floor-exclusive-max-100",
                "max_value": 100.5,
                "strict_max": True,
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "floor-exclusive-max-100"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "gear-exclusive-0-6"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "gear-exclusive-0-6",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {
                "column": "gear-exclusive-0-6",
                "min_value": 0.5,
                "strict_min": True,
                "max_value": 6.5,
                "strict_max": True,
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "gear-exclusive-0-6"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "optional-min-half"},
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "optional-min-half",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
            "meta": {},
        },
        {
            "expectation_type": "expect_column_values_to_be_between",
            "kwargs": {"column": "optional-min-half", "min_value": 0.5},
            "meta": {},
        },
    ]
    # fmt: on
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_has_profile_create_expectations_from_complex_schema(
    empty_data_context, complex_flat_schema
):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(complex_flat_schema, "complex")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "complex"
    assert obs.meta["notes"] == {
        "format": MetaNotesFormat.MARKDOWN,
        "content": ["### Description:\nAn address"],
    }

    # fmt: off
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "post-office-box"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "post-office-box",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "post-office-box"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "street-name"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "street-name",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "street-name"},
        },
        {
            "meta": {
                "notes": {
                    "format": MetaNotesFormat.MARKDOWN,
                    "content": ["### Description:\nOnly the address number."],
                }
            },
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "street-number"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "street-number",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "street-number"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "locality"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "locality",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "locality"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "region"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "region",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "region"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "postal-code"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "postal-code",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "postal-code"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_to_exist",
            "kwargs": {"column": "country-name"},
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "country-name",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "meta": {},
            "expectation_type": "expect_column_values_to_not_be_null",
            "kwargs": {"column": "country-name"},
        },
    ]
    # fmt: on
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)


def test_null_fields_schema(empty_data_context, null_fields_schema):
    profiler = JsonSchemaProfiler()
    obs = profiler.profile(null_fields_schema, "null_fields")
    assert isinstance(obs, ExpectationSuite)
    assert obs.expectation_suite_name == "null_fields"
    # fmt: off
    assert [e.to_json_dict() for e in obs.expectations] == [
        {
            "expectation_type": "expect_column_to_exist",
            "meta": {},
            "kwargs": {"column": "null"},
        },
        {
            "expectation_type": "expect_column_values_to_be_null",
            "meta": {},
            "kwargs": {"column": "null"},
        },
        {
            "expectation_type": "expect_column_to_exist",
            "meta": {},
            "kwargs": {"column": "string-or-null"},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
            "kwargs": {
                "column": "string-or-null",
                "type_list": [
                    "CHAR",
                    "NCHAR",
                    "NTEXT",
                    "NVARCHAR",
                    "STRING",
                    "StringType",
                    "StringType()",
                    "TEXT",
                    "VARCHAR",
                    "dtype('O')",
                    "object",
                    "str",
                    "string",
                ],
            },
        },
        {
            "expectation_type": "expect_column_to_exist",
            "meta": {},
            "kwargs": {"column": "int-or-null"},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
            "kwargs": {
                "column": "int-or-null",
                "type_list": [
                    "BIGINT",
                    "BYTEINT",
                    "ByteType()",
                    "INT",
                    "INT64",
                    "INTEGER",
                    "Int16Dtype",
                    "Int32Dtype",
                    "Int64Dtype",
                    "Int8Dtype",
                    "IntegerType",
                    "IntegerType()",
                    "LongType",
                    "LongType()",
                    "SMALLINT",
                    "ShortType()",
                    "TINYINT",
                    "UInt16Dtype",
                    "UInt32Dtype",
                    "UInt64Dtype",
                    "UInt8Dtype",
                    "int",
                    "int16",
                    "int32",
                    "int64",
                    "int8",
                    "int_",
                    "integer",
                    "uint16",
                    "uint32",
                    "uint64",
                    "uint8",
                ],
            },
        },
        {
            "expectation_type": "expect_column_to_exist",
            "meta": {},
            "kwargs": {"column": "number-or-null"},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "meta": {},
            "kwargs": {
                "column": "number-or-null",
                "type_list": [
                    "DECIMAL",
                    "DOUBLE",
                    "DOUBLE_PRECISION",
                    "DecimalType()",
                    "DoubleType",
                    "DoubleType()",
                    "FLOAT",
                    "FLOAT4",
                    "FLOAT64",
                    "FLOAT8",
                    "FloatType",
                    "FloatType()",
                    "NUMERIC",
                    "REAL",
                    "float",
                    "float16",
                    "float32",
                    "float64",
                    "float_",
                    "number",
                ],
            },
        },
        {
            "expectation_type": "expect_column_to_exist",
            "meta": {},
            "kwargs": {"column": "enum-or-null"},
        },
        {
            "expectation_type": "expect_column_values_to_be_in_set",
            "meta": {},
            "kwargs": {"column": "enum-or-null", "value_set": ["a", "b", "c"]},
        },
    ]
    # fmt: on
    context = empty_data_context
    context.add_expectation_suite(expectation_suite=obs)



import os
import shutil

import nbformat
import pytest

import great_expectations as gx
from great_expectations import DataContext
from great_expectations.data_context.util import file_relative_path
from great_expectations.render.renderer.checkpoint_new_notebook_renderer import (
    CheckpointNewNotebookRenderer,
)
from great_expectations.util import get_context


@pytest.fixture
def assetless_dataconnector_context(
    tmp_path_factory,
    monkeypatch,
):
    # Re-enable GE_USAGE_STATS
    monkeypatch.delenv("GE_USAGE_STATS")

    project_path = str(tmp_path_factory.mktemp("titanic_data_context"))
    context_path = os.path.join(project_path, "great_expectations")
    os.makedirs(os.path.join(context_path, "expectations"), exist_ok=True)
    data_path = os.path.join(context_path, "..", "data", "titanic")
    os.makedirs(os.path.join(data_path), exist_ok=True)
    shutil.copy(
        file_relative_path(
            __file__,
            "../../test_fixtures/great_expectations_v013_no_datasource_stats_enabled.yml",
        ),
        str(os.path.join(context_path, "great_expectations.yml")),
    )
    context = gx.get_context(context_root_dir=context_path)
    assert context.root_directory == context_path

    datasource_config = f"""
            class_name: Datasource

            execution_engine:
                class_name: PandasExecutionEngine

            data_connectors:
                my_other_data_connector:
                    class_name: ConfiguredAssetFilesystemDataConnector
                    base_directory: {data_path}
                    glob_directive: "*.csv"

                    default_regex:
                        pattern: (.+)\\.csv
                        group_names:
                            - name
                    assets:
                        {{}}
            """

    context.test_yaml_config(
        name="my_datasource", yaml_config=datasource_config, pretty_print=False
    )
    # noinspection PyProtectedMember
    context._save_project_config()
    return context


def test_find_datasource_with_asset_on_context_with_no_datasources(
    empty_data_context,
):
    context = empty_data_context
    assert len(context.list_datasources()) == 0

    renderer = CheckpointNewNotebookRenderer(context, "foo")
    obs = renderer._find_datasource_with_asset()
    assert obs is None


def test_find_datasource_with_asset_on_context_with_a_datasource_with_no_dataconnectors(
    titanic_pandas_data_context_with_v013_datasource_stats_enabled_with_checkpoints_v1_with_templates,
):
    context = titanic_pandas_data_context_with_v013_datasource_stats_enabled_with_checkpoints_v1_with_templates
    context.delete_datasource("my_datasource")
    assert len(context.list_datasources()) == 0
    context.add_datasource(
        "aaa_datasource",
        class_name="Datasource",
        module_name="great_expectations.datasource.new_datasource",
        execution_engine={
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
    )
    assert len(context.list_datasources()) == 1

    renderer = CheckpointNewNotebookRenderer(context, "foo")
    obs = renderer._find_datasource_with_asset()
    assert obs is None


@pytest.mark.slow  # 2.27s
def test_find_datasource_with_asset_on_context_with_a_datasource_with_a_dataconnector_that_has_no_assets(
    assetless_dataconnector_context,
):
    context = assetless_dataconnector_context
    assert list(context.get_datasource("my_datasource").data_connectors.keys()) == [
        "my_other_data_connector"
    ]

    # remove data asset name
    config = context.get_config_with_variables_substituted()
    root_directory = context.root_directory

    context = get_context(project_config=config, context_root_dir=root_directory)

    renderer = CheckpointNewNotebookRenderer(context, "foo")
    obs = renderer._find_datasource_with_asset()
    assert obs is None


def test_find_datasource_with_asset_on_happy_path_context(
    deterministic_asset_dataconnector_context,
):
    context = deterministic_asset_dataconnector_context
    assert len(context.list_datasources()) == 1

    renderer = CheckpointNewNotebookRenderer(context, "foo")
    obs = renderer._find_datasource_with_asset()

    assert obs == {
        "asset_name": "users",
        "data_connector_name": "my_other_data_connector",
        "datasource_name": "my_datasource",
    }


def test_find_datasource_with_asset_on_context_with_a_full_datasource_and_one_with_no_dataconnectors(
    deterministic_asset_dataconnector_context,
):
    context = deterministic_asset_dataconnector_context
    assert len(context.list_datasources()) == 1
    context.add_datasource(
        "aaa_datasource",
        class_name="Datasource",
        module_name="great_expectations.datasource.new_datasource",
        execution_engine={
            "class_name": "PandasExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
    )
    assert len(context.list_datasources()) == 2

    renderer = CheckpointNewNotebookRenderer(context, "foo")
    obs = renderer._find_datasource_with_asset()
    assert obs == {
        "datasource_name": "my_datasource",
        "data_connector_name": "my_other_data_connector",
        "asset_name": "users",
    }


@pytest.fixture
def checkpoint_new_notebook_assets():
    header = [
        {
            "cell_type": "markdown",
            "source": "# Create Your Checkpoint\nUse this notebook to configure a new Checkpoint and add it to your project:\n\n**Checkpoint Name**: `my_checkpoint_name`",
            "metadata": {},
        }
    ]
    imports = [
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": "from ruamel.yaml import YAML\nimport great_expectations as gx\nfrom pprint import pprint\n\nyaml = YAML()\ncontext = gx.get_context()",
            "outputs": [],
        },
    ]
    optional_customize_your_config = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Customize Your Configuration\nThe following cells show examples for listing your current configuration. You can replace values in the sample configuration with these values to customize your Checkpoint.""",
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": """# Run this cell to print out the names of your Datasources, Data Connectors and Data Assets\npprint(context.get_available_data_asset_names())""",
            "outputs": [],
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": "context.list_expectation_suite_names()",
            "outputs": [],
        },
    ]

    sample_checkpoint_config_markdown_description = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Create a Checkpoint Configuration\n\n**If you are new to Great Expectations or the Checkpoint feature**, you should start with SimpleCheckpoint because it includes default configurations like a default list of post validation actions.\n\nIn the cell below we have created a sample Checkpoint configuration using **your configuration** and **SimpleCheckpoint** to run a single validation of a single Expectation Suite against a single Batch of data.\n\nTo keep it simple, we are just choosing the first available instance of each of the following items you have configured in your Data Context:\n* Datasource\n* DataConnector\n* DataAsset\n* Partition\n* Expectation Suite\n\nOf course this is purely an example, you may edit this to your heart's content.\n\n**My configuration is not so simple - are there more advanced options?**\n\nGlad you asked! Checkpoints are very versatile. For example, you can validate many Batches in a single Checkpoint, validate Batches against different Expectation Suites or against many Expectation Suites, control the specific post-validation actions based on Expectation Suite / Batch / results of validation among other features. Check out our documentation on Checkpoints for more details and for instructions on how to implement other more advanced features including using the **Checkpoint** class:\n- https://docs.greatexpectations.io/docs/reference/checkpoints_and_actions\n- https://docs.greatexpectations.io/docs/guides/validation/checkpoints/how_to_create_a_new_checkpoint\n- https://docs.greatexpectations.io/docs/guides/validation/checkpoints/how_to_configure_a_new_checkpoint_using_test_yaml_config""",
        },
    ]
    sample_checkpoint_config_code_correct = [
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": (
                'my_checkpoint_name = "my_checkpoint_name"  # This was populated from your CLI command.\n\n'
                'yaml_config = f"""\n'
                "name: {my_checkpoint_name}\n"
                """config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
validations:
  - batch_request:
      datasource_name: my_datasource
      data_connector_name: my_other_data_connector
      data_asset_name: users
      data_connector_query:
        index: -1
    expectation_suite_name: Titanic.warning
"""
                '"""'
                "\nprint(yaml_config)"
            ),
            "outputs": [],
        },
    ]
    sample_checkpoint_config_markdown_error_message = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": "Sorry, we were unable to create a sample configuration. Perhaps you don't have a Datasource or Expectation Suite configured.",
        },
    ]

    test_and_save_your_checkpoint_configuration = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Test Your Checkpoint Configuration\nHere we will test your Checkpoint configuration to make sure it is valid.\n\nThis `test_yaml_config()` function is meant to enable fast dev loops. If your configuration is correct, this cell will show a message that you successfully instantiated a Checkpoint. You can continually edit your Checkpoint config yaml and re-run the cell to check until the new config is valid.\n\nIf you instead wish to use python instead of yaml to configure your Checkpoint, you can use `context.add_checkpoint()` and specify all the required parameters.""",
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": """my_checkpoint = context.test_yaml_config(yaml_config=yaml_config)""",
            "outputs": [],
        },
    ]
    review_checkpoint = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Review Your Checkpoint\n\nYou can run the following cell to print out the full yaml configuration. For example, if you used **SimpleCheckpoint**  this will show you the default action list.""",
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": 'print(my_checkpoint.get_config(mode="yaml"))',
            "outputs": [],
        },
    ]
    add_checkpoint = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Add Your Checkpoint\n\nRun the following cell to save this Checkpoint to your Checkpoint Store.""",
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": "context.add_checkpoint(**yaml.load(yaml_config))",
            "outputs": [],
        },
    ]
    optional_run_checkpoint = [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": """# Run Your Checkpoint & Open Data Docs(Optional)\n\nYou may wish to run the Checkpoint now and review its output in Data Docs. If so uncomment and run the following cell.""",
        },
        {
            "cell_type": "code",
            "metadata": {},
            "execution_count": None,
            "source": "# context.run_checkpoint(checkpoint_name=my_checkpoint_name)\n# context.open_data_docs()",
            "outputs": [],
        },
    ]

    return {
        "header": header,
        "imports": imports,
        "optional_customize_your_config": optional_customize_your_config,
        "sample_checkpoint_config_markdown_description": sample_checkpoint_config_markdown_description,
        "sample_checkpoint_config_code_correct": sample_checkpoint_config_code_correct,
        "sample_checkpoint_config_markdown_error_message": sample_checkpoint_config_markdown_error_message,
        "test_and_save_your_checkpoint_configuration": test_and_save_your_checkpoint_configuration,
        "review_checkpoint": review_checkpoint,
        "add_checkpoint": add_checkpoint,
        "optional_run_checkpoint": optional_run_checkpoint,
    }


@pytest.mark.slow  # 1.10s
def test_render_checkpoint_new_notebook_with_available_data_asset(
    deterministic_asset_dataconnector_context,
    titanic_expectation_suite,
    checkpoint_new_notebook_assets,
):
    """
    What does this test and why?
    The CheckpointNewNotebookRenderer should generate a notebook with an example SimpleCheckpoint yaml config based on the first available data asset.
    """

    context: DataContext = deterministic_asset_dataconnector_context

    assert context.list_checkpoints() == []
    context.add_expectation_suite(expectation_suite=titanic_expectation_suite)
    assert context.list_expectation_suite_names() == ["Titanic.warning"]

    checkpoint_new_notebook_renderer = CheckpointNewNotebookRenderer(
        context=context, checkpoint_name="my_checkpoint_name"
    )
    obs: nbformat.NotebookNode = checkpoint_new_notebook_renderer.render()

    assert isinstance(obs, dict)

    expected_cells = (
        checkpoint_new_notebook_assets["header"]
        + checkpoint_new_notebook_assets["imports"]
        + checkpoint_new_notebook_assets[
            "sample_checkpoint_config_markdown_description"
        ]
        # Testing to make sure everything in the notebook but especially this checkpoint config code is correct.
        + checkpoint_new_notebook_assets["sample_checkpoint_config_code_correct"]
        + checkpoint_new_notebook_assets["optional_customize_your_config"]
        + checkpoint_new_notebook_assets["test_and_save_your_checkpoint_configuration"]
        + checkpoint_new_notebook_assets["review_checkpoint"]
        + checkpoint_new_notebook_assets["add_checkpoint"]
        + checkpoint_new_notebook_assets["optional_run_checkpoint"]
    )

    expected = {
        "nbformat": 4,
        "nbformat_minor": 4,
        "metadata": {},
        "cells": expected_cells,
    }

    del expected["nbformat_minor"]
    del obs["nbformat_minor"]
    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        obs_cell.pop("id", None)
        assert obs_cell == expected_cell
    assert obs == expected


def test_render_checkpoint_new_notebook_with_unavailable_data_asset(
    assetless_dataconnector_context,
    checkpoint_new_notebook_assets,
):
    context: DataContext = assetless_dataconnector_context

    assert context.list_checkpoints() == []

    # This config is bad because of a missing expectation suite

    checkpoint_new_notebook_renderer = CheckpointNewNotebookRenderer(
        context=context, checkpoint_name="my_checkpoint_name"
    )
    obs: nbformat.NotebookNode = checkpoint_new_notebook_renderer.render()

    assert isinstance(obs, dict)

    expected_cells = (
        checkpoint_new_notebook_assets["header"]
        + checkpoint_new_notebook_assets["imports"]
        + checkpoint_new_notebook_assets[
            "sample_checkpoint_config_markdown_description"
        ]
        # Testing to make sure the error message here is displayed appropriately
        + checkpoint_new_notebook_assets[
            "sample_checkpoint_config_markdown_error_message"
        ]
        + checkpoint_new_notebook_assets["optional_customize_your_config"]
        + checkpoint_new_notebook_assets["test_and_save_your_checkpoint_configuration"]
        + checkpoint_new_notebook_assets["review_checkpoint"]
        + checkpoint_new_notebook_assets["add_checkpoint"]
        + checkpoint_new_notebook_assets["optional_run_checkpoint"]
    )

    expected = {
        "nbformat": 4,
        "nbformat_minor": 4,
        "metadata": {},
        "cells": expected_cells,
    }

    del expected["nbformat_minor"]
    del obs["nbformat_minor"]
    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        obs_cell.pop("id", None)
        assert obs_cell == expected_cell
    assert obs == expected



# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py imports">
from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import BatchRequest

# </snippet>
from great_expectations.expectations.metrics.import_manager import sa

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py get_context">
context = gx.get_context()
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py datasource_yaml">
datasource_yaml = f"""
name: taxi_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

introspection:  # Each key in the "introspection" section is an InferredAssetSqlDataConnector
    whole_table:
        include_schema_name: true
        introspection_directives:
            include_views: true
        skip_inapplicable_tables: true  # skip and continue upon encountering introspection errors
        excluded_tables:  # a list of tables to ignore when inferring data asset_names
            - main.yellow_tripdata_sample_2019_03  # format: schema_name.table_name

    daily:
        introspection_directives:
            include_views: true
        skip_inapplicable_tables: true  # skip and continue upon encountering introspection errors
        include_schema_name: true
        included_tables:  # only include tables in this list when inferring data asset_names
            - main.yellow_tripdata_sample_2019_01  # format: schema_name.table_name
        splitter_method: split_on_converted_datetime
        splitter_kwargs:
            column_name: pickup_datetime
            date_format_string: "%Y-%m-%d"

    hourly:
        introspection_directives:
            include_views: true
        skip_inapplicable_tables: true
        include_schema_name: true
        included_tables:  # only include tables in this list when inferring data asset_names
            - main.yellow_tripdata_sample_2019_01  # format: schema_name.table_name
        splitter_method: split_on_converted_datetime
        splitter_kwargs:
            column_name: pickup_datetime
            date_format_string: "%Y-%m-%d %H"

tables:  # Each key in the "tables" section is a table_name (key name "tables" in "SimpleSqlalchemyDatasource" configuration is reserved).
    # data_asset_name is: concatenate(data_asset_name_prefix, schema_name, table_name, data_asset_name_suffix)
    yellow_tripdata_sample_2019_01:  # Must match table name exactly.
        partitioners:  # Each key in the "partitioners" sub-section the name of a ConfiguredAssetSqlDataConnector (key name "partitioners" in "SimpleSqlalchemyDatasource" configuration is reserved).
            whole_table:
                include_schema_name: true
                schema_name: main
                data_asset_name_prefix: taxi__
                data_asset_name_suffix: __asset

            by_num_riders:
                include_schema_name: true
                schema_name: main
                data_asset_name_prefix: taxi__
                data_asset_name_suffix: __asset
                splitter_method: split_on_column_value
                splitter_kwargs:
                    column_name: passenger_count

            by_num_riders_random_sample:
                include_schema_name: true
                schema_name: main
                data_asset_name_prefix: taxi__
                data_asset_name_suffix: __asset
                splitter_method: split_on_column_value
                splitter_kwargs:
                    column_name: passenger_count
                sampling_method: _sample_using_random
                sampling_kwargs:
                    p: 1.0e-1
"""
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
data_dir_path = "data"
CONNECTION_STRING = f"sqlite:///{data_dir_path}/yellow_tripdata.db"

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py test_yaml_config">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py add_datasource">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>
available_data_asset_names = context.datasources[
    "taxi_datasource"
].get_available_data_asset_names(data_connector_names="whole_table")["whole_table"]
assert len(available_data_asset_names) == 3
# Here is a BatchRequest referring to an un-partitioned inferred data_asset.
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="whole_table",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name: str = "main.yellow_tripdata_sample_2019_01"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head(n_rows=10))

batch_list = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 1
batch_data = batch_list[0].data
num_rows = batch_data.execution_engine.engine.execute(
    sa.select([sa.func.count()]).select_from(batch_data.selectable)
).one()[0]
assert num_rows == 10000

# Here is a BatchRequest naming an inferred data_asset partitioned by day.
# This BatchRequest specifies multiple batches, which is useful for dataset exploration.
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="daily",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name: str = "main.yellow_tripdata_sample_2019_01"

batch_list = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 31  # number of days in January

# Here is a BatchRequest naming an inferred data_asset partitioned by hour.
# This BatchRequest specifies multiple batches, which is useful for dataset exploration.
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="hourly",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name: str = "main.yellow_tripdata_sample_2019_01"
batch_list = context.get_batch_list(batch_request=batch_request)
assert 24 <= len(batch_list) <= 744  # number of hours in a 31-day month

# Here is a BatchRequest naming a configured data_asset partitioned by passenger_count.
# This BatchRequest specifies multiple batches, which is useful for dataset exploration and data analysis.
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="by_num_riders",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name and other arguments directly in the BatchRequest above.
batch_request.data_asset_name: str = "taxi__main.yellow_tripdata_sample_2019_01__asset"

batch_list = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 6

# Here is a BatchRequest naming a configured data_asset partitioned by passenger_count.
# In addition, a randomly sampled fraction of each partition's batch data is obtained.
# Finally, a randomly sampled fraction of each partition's batch data is selected and returned as the final batch list.
# This BatchRequest specifies multiple batches, which is useful for dataset exploration and data analysis.
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py batch_request">
batch_request = BatchRequest(
    datasource_name="taxi_datasource",
    data_connector_name="by_num_riders_random_sample",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name and other arguments directly in the BatchRequest above.
batch_request.data_asset_name: str = "taxi__main.yellow_tripdata_sample_2019_01__asset"

batch_list = context.get_batch_list(batch_request=batch_request)
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_complete.py assertions">
assert len(batch_list) == 6  # ride occupancy ranges from 1 passenger to 6 passengers

batch_data = batch_list[1].data  # 2-passenger sample of batch data
num_rows = batch_data.execution_engine.engine.execute(
    sa.select([sa.func.count()]).select_from(batch_data.selectable)
).scalar()
assert num_rows < 200
# </snippet>

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert "taxi_datasource" in [ds["name"] for ds in context.list_datasources()]
assert "main.yellow_tripdata_sample_2019_01" in set(
    context.get_available_data_asset_names()["taxi_datasource"]["whole_table"]
)
assert "taxi__main.yellow_tripdata_sample_2019_01__asset" in set(
    context.get_available_data_asset_names()["taxi_datasource"][
        "by_num_riders_random_sample"
    ]
)



import os
import shutil

import pytest

from great_expectations.data_context.util import file_relative_path


@pytest.fixture
def empty_context_with_checkpoint_v1_stats_enabled(
    empty_data_context_stats_enabled, monkeypatch
):
    try:
        monkeypatch.delenv("VAR")
        monkeypatch.delenv("MY_PARAM")
        monkeypatch.delenv("OLD_PARAM")
    except:
        pass

    monkeypatch.setenv("VAR", "test")
    monkeypatch.setenv("MY_PARAM", "1")
    monkeypatch.setenv("OLD_PARAM", "2")

    context = empty_data_context_stats_enabled
    root_dir = context.root_directory
    fixture_name = "my_v1_checkpoint.yml"
    fixture_path = file_relative_path(
        __file__, f"../data_context/fixtures/contexts/{fixture_name}"
    )
    checkpoints_file = os.path.join(root_dir, "checkpoints", fixture_name)
    shutil.copy(fixture_path, checkpoints_file)
    # # noinspection PyProtectedMember
    context._save_project_config()
    return context


@pytest.fixture
def v10_project_directory(tmp_path_factory):
    """
    GX 0.10.x project for testing upgrade helper
    """
    project_path = str(tmp_path_factory.mktemp("v10_project"))
    context_root_dir = os.path.join(project_path, "great_expectations")
    shutil.copytree(
        file_relative_path(
            __file__, "../test_fixtures/upgrade_helper/great_expectations_v10_project/"
        ),
        context_root_dir,
    )
    shutil.copy(
        file_relative_path(
            __file__, "../test_fixtures/upgrade_helper/great_expectations_v1_basic.yml"
        ),
        os.path.join(context_root_dir, "great_expectations.yml"),
    )
    return context_root_dir


@pytest.fixture(scope="function")
def misc_directory(tmp_path):
    misc_dir = tmp_path / "random"
    misc_dir.mkdir()
    assert os.path.isabs(misc_dir)
    return misc_dir



import json
import logging
from typing import Any

from great_expectations.core.batch import BatchDefinition  # noqa: TCH001
from great_expectations.datasource.data_connector.sorter import Sorter

logger = logging.getLogger(__name__)


class LexicographicSorter(Sorter):
    def get_batch_key(self, batch_definition: BatchDefinition) -> Any:
        batch_identifiers: dict = batch_definition.batch_identifiers
        batch_value: Any = batch_identifiers[self.name]
        return batch_value

    def __repr__(self) -> str:
        doc_fields_dict: dict = {
            "name": self.name,
            "reverse": self.reverse,
            "type": "LexicographicSorter",
        }
        return json.dumps(doc_fields_dict, indent=2)



import random

import pytest
from ruamel.yaml import YAML

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch import Batch, BatchDefinition
from great_expectations.core.id_dict import IDDict
from great_expectations.datasource import Datasource
from great_expectations.datasource.data_connector import (
    InferredAssetAWSGlueDataCatalogDataConnector,
)
from great_expectations.validator.validator import Validator

yaml = YAML(typ="safe")


def test_basic_instantiation(glue_titanic_catalog):
    random.seed(0)

    config = yaml.load(
        """
    name: my_data_connector
    datasource_name: FAKE_DATASOURCE_NAME
    execution_engine: execution_engine
    """,
    )
    my_data_connector = InferredAssetAWSGlueDataCatalogDataConnector(**config)

    # noinspection PyProtectedMember
    asset_names = my_data_connector.get_available_data_asset_names()

    assert len(asset_names) == 2
    assert (
        "db_test.tb_titanic_with_partitions" in asset_names
        and "db_test.tb_titanic_without_partitions" in asset_names
    )
    assert my_data_connector.get_unmatched_data_references() == []

    report = my_data_connector.self_check()
    assert report == {
        "class_name": "InferredAssetAWSGlueDataCatalogDataConnector",
        "data_asset_count": 2,
        "example_data_asset_names": [
            "db_test.tb_titanic_with_partitions",
            "db_test.tb_titanic_without_partitions",
        ],
        "data_assets": {
            "db_test.tb_titanic_with_partitions": {
                "batch_definition_count": 6,
                "example_data_references": [
                    {"PClass": "1st", "SexCode": "0"},
                    {"PClass": "1st", "SexCode": "1"},
                    {"PClass": "2nd", "SexCode": "0"},
                ],
            },
            "db_test.tb_titanic_without_partitions": {
                "batch_definition_count": 1,
                "example_data_references": [{}],
            },
        },
        "unmatched_data_reference_count": 0,
        "example_unmatched_data_references": [],
    }
    assert "db_test.tb_titanic_with_partitions" in my_data_connector.assets
    assert my_data_connector.assets["db_test.tb_titanic_with_partitions"] == {
        "database_name": "db_test",
        "table_name": "tb_titanic_with_partitions",
        "partitions": ["PClass", "SexCode"],
        "data_asset_name_prefix": "",
        "data_asset_name_suffix": "",
    }
    assert "db_test.tb_titanic_without_partitions" in my_data_connector.assets
    assert my_data_connector.assets["db_test.tb_titanic_without_partitions"] == {
        "database_name": "db_test",
        "table_name": "tb_titanic_without_partitions",
        "partitions": [],
        "data_asset_name_prefix": "",
        "data_asset_name_suffix": "",
    }


def test_instantiation_from_a_config(
    glue_titanic_catalog, empty_data_context_stats_enabled
):
    random.seed(0)
    report_object = empty_data_context_stats_enabled.test_yaml_config(
        f"""
        module_name: great_expectations.datasource.data_connector
        class_name: InferredAssetAWSGlueDataCatalogDataConnector
        name: my_data_connector
        datasource_name: FAKE_Datasource_NAME
        data_asset_name_prefix: prefix__
        data_asset_name_suffix: __suffix
        excluded_tables:
            - db_test.tb_titanic_without_partitions
        glue_introspection_directives:
            database_name: db_test
        """,
        runtime_environment={
            "execution_engine": "execution_engine",
        },
        return_mode="report_object",
    )

    assert report_object == {
        "class_name": "InferredAssetAWSGlueDataCatalogDataConnector",
        "data_asset_count": 1,
        "example_data_asset_names": [
            "prefix__db_test.tb_titanic_with_partitions__suffix",
        ],
        "data_assets": {
            "prefix__db_test.tb_titanic_with_partitions__suffix": {
                "batch_definition_count": 6,
                "example_data_references": [
                    {"PClass": "1st", "SexCode": "0"},
                    {"PClass": "1st", "SexCode": "1"},
                    {"PClass": "2nd", "SexCode": "0"},
                ],
            },
        },
        "unmatched_data_reference_count": 0,
        "example_unmatched_data_references": [],
    }


def test_instantiation_with_included_tables(glue_titanic_catalog):
    my_data_connector = InferredAssetAWSGlueDataCatalogDataConnector(
        name="my_data_connector",
        datasource_name="FAKE_DATASOURCE_NAME",
        execution_engine="execution_engine",
        catalog_id="catalog_A",
        included_tables=["db_test.tb_titanic_with_partitions"],
    )

    assert len(my_data_connector.assets) == 1
    assert "db_test.tb_titanic_with_partitions" in my_data_connector.assets


def test_instantiation_with_invalid_database_name(glue_titanic_catalog):
    with pytest.raises(gx_exceptions.DataConnectorError):
        InferredAssetAWSGlueDataCatalogDataConnector(
            name="my_data_connector",
            datasource_name="FAKE_DATASOURCE_NAME",
            execution_engine="execution_engine",
            catalog_id="catalog_A",
            glue_introspection_directives={"database_name": "FAKE_DATABASE"},
        )


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_without_partitions(
    splitter_method_name_prefix,
    glue_titanic_catalog,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    my_data_connector = InferredAssetAWSGlueDataCatalogDataConnector(
        name="my_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine=test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
        catalog_id="catalog_A",
        data_asset_name_prefix="prefix__",
        data_asset_name_suffix="__suffix",
    )
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=BatchDefinition(
            datasource_name="FAKE_Datasource_NAME",
            data_connector_name="my_data_connector",
            data_asset_name="prefix__db_test.tb_titanic_without_partitions__suffix",
            batch_identifiers=IDDict(),
            batch_spec_passthrough={
                "splitter_method": f"{splitter_method_name_prefix}split_on_hashed_column",
                "splitter_kwargs": {
                    "column_name": "Name",
                    "hash_digits": 1,
                    "hash_function_name": "md5",
                    "batch_identifiers": {"hash_value": "f"},
                },
            },
        )
    )

    batch = Batch(data=batch_data)

    validator = Validator(execution_engine, batches=[batch])

    assert len(validator.head(fetch_all=True)) == 77


def test_get_batch_data_and_metadata_with_partitions(
    in_memory_runtime_context,
    glue_titanic_catalog,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_data_connector": {
                "class_name": "InferredAssetAWSGlueDataCatalogDataConnector",
                "catalog_id": "catalog_A",
                "data_asset_name_prefix": "prefix__",
                "data_asset_name_suffix": "__suffix",
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_data_connector"]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_data_connector",
        data_asset_name="prefix__db_test.tb_titanic_with_partitions__suffix",
        batch_identifiers=IDDict({"PClass": "1st", "SexCode": "0"}),
    )
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )

    batch = Batch(data=batch_data, batch_definition=batch_definition)

    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )

    assert validator.expect_column_values_to_be_in_set("PClass", ["1st"]).success
    assert validator.expect_column_values_to_be_in_set("SexCode", ["0"]).success



import os

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py imports">
import great_expectations as gx
from great_expectations.core.batch import BatchRequest

# </snippet>
from great_expectations.core.yaml_handler import YAMLHandler

yaml = YAMLHandler()

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py get_context">
context = gx.get_context()
# </snippet>


# NOTE: The following code is only for testing and depends on an environment
# variable to set the gcp_project. You can replace the value with your own
# GCP project information
gcp_project = os.environ.get("GE_TEST_GCP_PROJECT")
if not gcp_project:
    raise ValueError(
        "Environment Variable GE_TEST_GCP_PROJECT is required to run GCS integration tests"
    )

# parse great_expectations.yml for comparison
great_expectations_yaml_file_path = os.path.join(
    context.root_directory, "great_expectations.yml"
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)

stores = great_expectations_yaml["stores"]
pop_stores = ["checkpoint_store", "evaluation_parameter_store", "validations_store"]
for store in pop_stores:
    stores.pop(store)

actual_existing_expectations_store = {}
actual_existing_expectations_store["stores"] = stores
actual_existing_expectations_store["expectations_store_name"] = great_expectations_yaml[
    "expectations_store_name"
]

expected_existing_expectations_store_yaml = """
stores:
  expectations_store:
    class_name: ExpectationsStore
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: expectations/

expectations_store_name: expectations_store
"""

assert actual_existing_expectations_store == yaml.load(
    expected_existing_expectations_store_yaml
)

# adding expectations store
configured_expectations_store_yaml = """
stores:
  expectations_GCS_store:
    class_name: ExpectationsStore
    store_backend:
      class_name: TupleGCSStoreBackend
      project: <YOUR GCP PROJECT NAME>
      bucket: <YOUR GCS BUCKET NAME>
      prefix: <YOUR GCS PREFIX NAME>

expectations_store_name: expectations_GCS_store
"""

# replace example code with integration test configuration
configured_expectations_store = yaml.load(configured_expectations_store_yaml)
configured_expectations_store["stores"]["expectations_GCS_store"]["store_backend"][
    "project"
] = gcp_project
configured_expectations_store["stores"]["expectations_GCS_store"]["store_backend"][
    "bucket"
] = "test_metadata_store"
configured_expectations_store["stores"]["expectations_GCS_store"]["store_backend"][
    "prefix"
] = "metadata/expectations"

# add and set the new expectation store
context.add_store(
    store_name=configured_expectations_store["expectations_store_name"],
    store_config=configured_expectations_store["stores"]["expectations_GCS_store"],
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)
great_expectations_yaml["expectations_store_name"] = "expectations_GCS_store"
great_expectations_yaml["stores"]["expectations_GCS_store"]["store_backend"].pop(
    "suppress_store_backend_id"
)
with open(great_expectations_yaml_file_path, "w") as f:
    yaml.dump(great_expectations_yaml, f)

# adding validation results store

# parse great_expectations.yml for comparison
great_expectations_yaml_file_path = os.path.join(
    context.root_directory, "great_expectations.yml"
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)

stores = great_expectations_yaml["stores"]
# popping the rest out so taht we can do the comparison. They aren't going anywhere dont worry
pop_stores = [
    "checkpoint_store",
    "evaluation_parameter_store",
    "expectations_store",
    "expectations_GCS_store",
]
for store in pop_stores:
    stores.pop(store)

actual_existing_validations_store = {}
actual_existing_validations_store["stores"] = stores
actual_existing_validations_store["validations_store_name"] = great_expectations_yaml[
    "validations_store_name"
]

expected_existing_validations_store_yaml = """
stores:
  validations_store:
    class_name: ValidationsStore
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: uncommitted/validations/

validations_store_name: validations_store
"""
assert actual_existing_validations_store == yaml.load(
    expected_existing_validations_store_yaml
)

# adding validations store
configured_validations_store_yaml = """
stores:
  validations_GCS_store:
    class_name: ValidationsStore
    store_backend:
      class_name: TupleGCSStoreBackend
      project: <YOUR GCP PROJECT NAME>
      bucket: <YOUR GCS BUCKET NAME>
      prefix: <YOUR GCS PREFIX NAME>

validations_store_name: validations_GCS_store
"""

# replace example code with integration test configuration
configured_validations_store = yaml.load(configured_validations_store_yaml)
configured_validations_store["stores"]["validations_GCS_store"]["store_backend"][
    "project"
] = gcp_project
configured_validations_store["stores"]["validations_GCS_store"]["store_backend"][
    "bucket"
] = "test_metadata_store"
configured_validations_store["stores"]["validations_GCS_store"]["store_backend"][
    "prefix"
] = "metadata/validations"

# add and set the new validation store
context.add_store(
    store_name=configured_validations_store["validations_store_name"],
    store_config=configured_validations_store["stores"]["validations_GCS_store"],
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)
great_expectations_yaml["validations_store_name"] = "validations_GCS_store"
great_expectations_yaml["stores"]["validations_GCS_store"]["store_backend"].pop(
    "suppress_store_backend_id"
)
with open(great_expectations_yaml_file_path, "w") as f:
    yaml.dump(great_expectations_yaml, f)

# adding data docs store
data_docs_site_yaml = """
data_docs_sites:
  local_site:
    class_name: SiteBuilder
    show_how_to_buttons: true
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: uncommitted/data_docs/local_site/
    site_index_builder:
      class_name: DefaultSiteIndexBuilder
  gs_site:  # this is a user-selected name - you may select your own
    class_name: SiteBuilder
    store_backend:
      class_name: TupleGCSStoreBackend
      project: <YOUR GCP PROJECT NAME>
      bucket: <YOUR GCS BUCKET NAME>
    site_index_builder:
      class_name: DefaultSiteIndexBuilder
"""
data_docs_site_yaml = data_docs_site_yaml.replace(
    "<YOUR GCP PROJECT NAME>", gcp_project
)
data_docs_site_yaml = data_docs_site_yaml.replace(
    "<YOUR GCS BUCKET NAME>", "test_datadocs_store"
)
great_expectations_yaml_file_path = os.path.join(
    context.root_directory, "great_expectations.yml"
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)
great_expectations_yaml["data_docs_sites"] = yaml.load(data_docs_site_yaml)[
    "data_docs_sites"
]
with open(great_expectations_yaml_file_path, "w") as f:
    yaml.dump(great_expectations_yaml, f)

# adding datasource
# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py datasource_yaml">
datasource_yaml = rf"""
name: my_gcs_datasource
class_name: Datasource
execution_engine:
    class_name: PandasExecutionEngine
data_connectors:
    default_runtime_data_connector_name:
        class_name: RuntimeDataConnector
        batch_identifiers:
            - default_identifier_name
    default_inferred_data_connector_name:
        class_name: InferredAssetGCSDataConnector
        bucket_or_name: <YOUR_GCS_BUCKET_HERE>
        prefix: <BUCKET_PATH_TO_DATA>
        default_regex:
            pattern: (.*)\.csv
            group_names:
                - data_asset_name
"""
# </snippet>


# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_yaml = datasource_yaml.replace("<YOUR_GCS_BUCKET_HERE>", "test_docs_data")
datasource_yaml = datasource_yaml.replace(
    "<BUCKET_PATH_TO_DATA>", "data/taxi_yellow_tripdata_samples/"
)

context.test_yaml_config(datasource_yaml)
# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py add_datasource">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>

# batch_request with data_asset_name
# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py batch_request">
batch_request = BatchRequest(
    datasource_name="my_gcs_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)
# </snippet>


# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name = (
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01"
)

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py add_expectation_suite">
context.add_or_update_expectation_suite(expectation_suite_name="test_gcs_suite")

validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_gcs_suite"
)
# </snippet>


# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_gcs_datasource"]
assert set(
    context.get_available_data_asset_names()["my_gcs_datasource"][
        "default_inferred_data_connector_name"
    ]
) == {
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01",
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-02",
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-03",
}

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py validator_calls">
validator.expect_column_values_to_not_be_null(column="passenger_count")

validator.expect_column_values_to_be_between(
    column="congestion_surcharge", min_value=0, max_value=1000
)
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py save_expectation_suite">
validator.save_expectation_suite(discard_failed_expectations=False)
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py checkpoint_config">
my_checkpoint_name = "gcs_checkpoint"
checkpoint_config = f"""
name: {my_checkpoint_name}
config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
validations:
  - batch_request:
      datasource_name: my_gcs_datasource
      data_connector_name: default_inferred_data_connector_name
      data_asset_name: <YOUR_DATA_ASSET_NAME>
    expectation_suite_name: test_gcs_suite
"""
# </snippet>

checkpoint_config = checkpoint_config.replace(
    "<YOUR_DATA_ASSET_NAME>",
    "data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_2019-01",
)

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py add_checkpoint">
context.add_or_update_checkpoint(**yaml.load(checkpoint_config))
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/gcp_deployment_patterns_file_gcs_yaml_configs.py run_checkpoint">
checkpoint_result = context.run_checkpoint(
    checkpoint_name=my_checkpoint_name,
)
# </snippet>

assert checkpoint_result.success is True



from typing import Dict, List, Optional

import numpy as np
import pytest

from great_expectations import DataContext
from great_expectations.core.domain import Domain
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import (
    SummarizationMetricNameSuffixes,
)
from great_expectations.rule_based_profiler.config import ParameterBuilderConfig
from great_expectations.rule_based_profiler.helpers.util import (
    get_parameter_value_and_validate_return_type,
)
from great_expectations.rule_based_profiler.parameter_builder import (
    MeanUnexpectedMapMetricMultiBatchParameterBuilder,
    MetricMultiBatchParameterBuilder,
    ParameterBuilder,
)
from great_expectations.rule_based_profiler.parameter_container import (
    DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
    ParameterContainer,
    ParameterNode,
)
from tests.rule_based_profiler.conftest import ATOL, RTOL


@pytest.mark.integration
def test_instantiation_mean_unexpected_map_metric_multi_batch_parameter_builder(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    # noinspection PyUnusedLocal
    parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_name",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            data_context=data_context,
        )
    )


@pytest.mark.integration
def test_instantiation_mean_unexpected_map_metric_multi_batch_parameter_builder_required_arguments_absent(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    with pytest.raises(TypeError) as excinfo:
        # noinspection PyUnusedLocal,PyArgumentList
        parameter_builder: ParameterBuilder = (
            MeanUnexpectedMapMetricMultiBatchParameterBuilder(
                name="my_name",
                map_metric_name="column_values.nonnull",
                data_context=data_context,
            )
        )

    assert (
        "__init__() missing 1 required positional argument: 'total_count_parameter_builder_name'"
        in str(excinfo.value)
    )

    with pytest.raises(TypeError) as excinfo:
        # noinspection PyUnusedLocal,PyArgumentList
        parameter_builder: ParameterBuilder = (
            MeanUnexpectedMapMetricMultiBatchParameterBuilder(
                name="my_name",
                total_count_parameter_builder_name="my_total_count",
                data_context=data_context,
            )
        )

    assert (
        "__init__() missing 1 required positional argument: 'map_metric_name'"
        in str(excinfo.value)
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.56s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_numeric_dependencies_evaluated_separately(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MetricMultiBatchParameterBuilder(
            name="my_total_count",
            metric_name="table.row_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            single_batch_mode=False,
            enforce_numeric_metric=False,
            replace_nan_with_zero=False,
            reduce_scalar_metric=True,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )
    my_null_count_metric_multi_batch_parameter_builder: ParameterBuilder = MetricMultiBatchParameterBuilder(
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        single_batch_mode=False,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
        data_context=data_context,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_passenger_count_values_not_null_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "passenger_count"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    my_total_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )
    my_null_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 0.0

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.58s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_numeric_dependencies_evaluated_in_parameter_builder(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_total_count",
        metric_name="table.row_count",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )
    my_null_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )

    evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
        my_total_count_metric_multi_batch_parameter_builder_config,
        my_null_count_metric_multi_batch_parameter_builder_config,
    ]
    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_passenger_count_values_not_null_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "passenger_count"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 0.0

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.58s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_numeric_dependencies_evaluated_mixed(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_total_count",
        metric_name="table.row_count",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )
    my_null_count_metric_multi_batch_parameter_builder: ParameterBuilder = MetricMultiBatchParameterBuilder(
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        single_batch_mode=False,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
        data_context=data_context,
    )

    evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
        my_total_count_metric_multi_batch_parameter_builder_config,
    ]
    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_passenger_count_values_not_null_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "passenger_count"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    my_null_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 0.0

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.58s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_datetime_dependencies_evaluated_separately(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MetricMultiBatchParameterBuilder(
            name="my_total_count",
            metric_name="table.row_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            single_batch_mode=False,
            enforce_numeric_metric=False,
            replace_nan_with_zero=False,
            reduce_scalar_metric=True,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )
    my_null_count_metric_multi_batch_parameter_builder: ParameterBuilder = MetricMultiBatchParameterBuilder(
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        single_batch_mode=False,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
        data_context=data_context,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_pickup_datetime_count_values_unique_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "pickup_datetime"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    my_total_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )
    my_null_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 3.89e-3

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.58s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_datetime_dependencies_evaluated_in_parameter_builder(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_total_count",
        metric_name="table.row_count",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )
    my_null_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )

    evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
        my_total_count_metric_multi_batch_parameter_builder_config,
        my_null_count_metric_multi_batch_parameter_builder_config,
    ]
    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_pickup_datetime_count_values_unique_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "pickup_datetime"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 3.89e-3

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
@pytest.mark.slow  # 1.65s
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_datetime_dependencies_evaluated_mixed(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    batch_request: dict = {
        "datasource_name": "taxi_pandas",
        "data_connector_name": "monthly",
        "data_asset_name": "my_reports",
    }

    my_total_count_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MetricMultiBatchParameterBuilder(
            name="my_total_count",
            metric_name="table.row_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            single_batch_mode=False,
            enforce_numeric_metric=False,
            replace_nan_with_zero=False,
            reduce_scalar_metric=True,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )
    my_null_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )

    evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
        my_null_count_metric_multi_batch_parameter_builder_config,
    ]
    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_pickup_datetime_count_values_unique_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )
    )

    metric_domain_kwargs: dict = {"column": "pickup_datetime"}
    domain = Domain(
        domain_type=MetricDomainTypes.COLUMN,
        domain_kwargs=metric_domain_kwargs,
        rule_name="my_rule",
    )

    variables: Optional[ParameterContainer] = None

    parameter_container = ParameterContainer(parameter_nodes=None)
    parameters: Dict[str, ParameterContainer] = {
        domain.id: parameter_container,
    }

    my_total_count_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder.build_parameters(
        domain=domain,
        variables=variables,
        parameters=parameters,
        batch_request=batch_request,
        runtime_configuration=None,
    )

    expected_parameter_value: float = 3.89e-3

    parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=mean_unexpected_map_metric_multi_batch_parameter_builder.json_serialized_fully_qualified_parameter_name,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )

    rtol: float = RTOL
    atol: float = 5.0e-1 * ATOL
    np.testing.assert_allclose(
        actual=parameter_node.value,
        desired=expected_parameter_value,
        rtol=rtol,
        atol=atol,
        err_msg=f"Actual value of {parameter_node.value} differs from expected value of {expected_parameter_value} by more than {atol + rtol * abs(parameter_node.value)} tolerance.",
    )


@pytest.mark.integration
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_check_serialized_keys_no_evaluation_parameter_builder_configs(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_pickup_datetime_count_values_unique_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=None,
            data_context=data_context,
        )
    )

    # Note: "evaluation_parameter_builder_configs" is not one of "ParameterBuilder" formal property attributes.
    assert set(
        mean_unexpected_map_metric_multi_batch_parameter_builder.to_json_dict().keys()
    ) == {
        "class_name",
        "module_name",
        "name",
        "map_metric_name",
        "total_count_parameter_builder_name",
        "null_count_parameter_builder_name",
        "metric_domain_kwargs",
        "metric_value_kwargs",
        "evaluation_parameter_builder_configs",
    }


@pytest.mark.integration
def test_mean_unexpected_map_metric_multi_batch_parameter_builder_bobby_check_serialized_keys_with_evaluation_parameter_builder_configs(
    bobby_columnar_table_multi_batch_deterministic_data_context,
):
    data_context: DataContext = (
        bobby_columnar_table_multi_batch_deterministic_data_context
    )

    my_total_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_total_count",
        metric_name="table.row_count",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )
    my_null_count_metric_multi_batch_parameter_builder_config = ParameterBuilderConfig(
        module_name="great_expectations.rule_based_profiler.parameter_builder",
        class_name="MetricMultiBatchParameterBuilder",
        name="my_null_count",
        metric_name=f"column_values.nonnull.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}",
        metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
        metric_value_kwargs=None,
        enforce_numeric_metric=False,
        replace_nan_with_zero=False,
        reduce_scalar_metric=True,
        evaluation_parameter_builder_configs=None,
    )

    evaluation_parameter_builder_configs: Optional[List[ParameterBuilderConfig]] = [
        my_total_count_metric_multi_batch_parameter_builder_config,
        my_null_count_metric_multi_batch_parameter_builder_config,
    ]
    mean_unexpected_map_metric_multi_batch_parameter_builder: ParameterBuilder = (
        MeanUnexpectedMapMetricMultiBatchParameterBuilder(
            name="my_pickup_datetime_count_values_unique_mean_unexpected_map_metric",
            map_metric_name="column_values.nonnull",
            total_count_parameter_builder_name="my_total_count",
            null_count_parameter_builder_name="my_null_count",
            metric_domain_kwargs=DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME,
            metric_value_kwargs=None,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )
    )

    # Note: "evaluation_parameter_builder_configs" is not one of "ParameterBuilder" formal property attributes.
    assert set(
        mean_unexpected_map_metric_multi_batch_parameter_builder.to_json_dict().keys()
    ) == {
        "class_name",
        "module_name",
        "name",
        "map_metric_name",
        "total_count_parameter_builder_name",
        "null_count_parameter_builder_name",
        "metric_domain_kwargs",
        "metric_value_kwargs",
        "evaluation_parameter_builder_configs",
    }
